// MilkshapeModel.c

#include "MilkshapeModel.h"
#include "Project.h"

BOOL Model_LoadModelData(MODEL * mod, const char *filename )
{

	char pathTemp[MAX_PATH+1];
	int pathLength;
	for ( pathLength = strlen( filename ); --pathLength; ) {
		if ( filename[pathLength] == '/' || filename[pathLength] == '\\' ) {
			break;
		}
	}
	strncpy( pathTemp, filename, pathLength );

	FILE * file = fopen( filename, "rb" );
	if ( !file )
		return FALSE;	// "Couldn't open the model file."

	fseek(file, 0, SEEK_END );
	long fileSize = ftell(file);
	fseek(file, 0, SEEK_SET );

	byte *pBuffer = malloc(fileSize);
	fread(pBuffer, fileSize, 1, file );
	fclose(file);

	const byte *pPtr = pBuffer;
	MS3DHEADER *pHeader = ( MS3DHEADER* )pPtr;
	pPtr += sizeof( MS3DHEADER );

	if ( strncmp( pHeader->ID, "MS3D000000", 10 ) != 0 )
		return FALSE; // "Not a valid Milkshape3D model file."

	if ( pHeader->version < 3 || pHeader->version > 4 )
		return FALSE; // "Unhandled file version. Only Milkshape3D Version 1.3 and 1.4 is supported." );

	int nVertices = *( word* )pPtr;
	mod->numVertices = nVertices;
	mod->pVertices   = malloc(sizeof(VERTEX) * nVertices);
	pPtr += sizeof( word );

	int i;
	for ( i = 0; i < nVertices; i++ )
	{
		MS3DVERTEX *pVertex = ( MS3DVERTEX* )pPtr;
		mod->pVertices[i].boneID = pVertex->boneID;
		memcpy( mod->pVertices[i].location, pVertex->vertex, sizeof( float )*3 );
		pPtr += sizeof( MS3DVERTEX );
	}

	int nTriangles = *( word* )pPtr;
	mod->numTriangles = nTriangles;
	mod->pTriangles   = malloc(sizeof(TRIANGLE) * nTriangles);
	pPtr += sizeof( word );
	int vertexIndices[3];
	float t[3];

	for ( i = 0; i < nTriangles; i++ )
	{
		MS3DTRIANGLE * pTriangle = ( MS3DTRIANGLE* )pPtr;
	   	vertexIndices[0] = pTriangle->vertexIndices[0];
		vertexIndices[1] = pTriangle->vertexIndices[1];
		vertexIndices[2] = pTriangle->vertexIndices[2];
		t[0] = 1.0f - pTriangle->t[0];
		t[1] = 1.0f - pTriangle->t[1];
		t[2] = 1.0f - pTriangle->t[2];
		memcpy( mod->pTriangles[i].vertexNormals, pTriangle->vertexNormals, sizeof( float )*3*3 );
		memcpy( mod->pTriangles[i].s, pTriangle->s, sizeof( float )*3 );
		memcpy( mod->pTriangles[i].t, t, sizeof( float )*3 );
		memcpy( mod->pTriangles[i].vertexIndices, vertexIndices, sizeof( int )*3 );
		pPtr += sizeof( MS3DTRIANGLE );
	}

	int nGroups = *( word* )pPtr;
	mod->numMeshes = nGroups;
	mod->pMeshes   = malloc(sizeof(MESH) * nGroups);
	pPtr += sizeof( word );
	for ( i = 0; i < nGroups; i++ )
	{
		pPtr += sizeof( byte );	// flags
		pPtr += 32;				// name

		nTriangles = *( word* )pPtr;
		pPtr += sizeof( word );
		int * pTriangleIndices = malloc(sizeof(int) * nTriangles);
		for ( int j = 0; j < nTriangles; j++ )
		{
			pTriangleIndices[j] = *( word* )pPtr;
			pPtr += sizeof( word );
		}

		char materialIndex = *( char* )pPtr;
		pPtr += sizeof( char );

		mod->pMeshes[i].materialIndex = materialIndex;
		mod->pMeshes[i].numTriangles = nTriangles;
		mod->pMeshes[i].pTriangleIndices = pTriangleIndices;
	}

	int nMaterials = *( word* )pPtr;
	mod->numMaterials = nMaterials;
	mod->pMaterials   = malloc(sizeof(MATERIAL) * nMaterials);
	pPtr += sizeof( word );
	for ( i = 0; i < nMaterials; i++ )
	{
		MS3DMATERIAL *pMaterial = ( MS3DMATERIAL* )pPtr;
		memcpy( mod->pMaterials[i].ambient, pMaterial->ambient, sizeof( float )*4 );
		memcpy( mod->pMaterials[i].diffuse, pMaterial->diffuse, sizeof( float )*4 );
		memcpy( mod->pMaterials[i].specular, pMaterial->specular, sizeof( float )*4 );
		memcpy( mod->pMaterials[i].emissive, pMaterial->emissive, sizeof( float )*4 );
		mod->pMaterials[i].shininess = pMaterial->shininess;

		if ( strncmp( pMaterial->texture, ".\\", 2 ) == 0 ) {
			// relative path
			strcpy( pathTemp + pathLength, pMaterial->texture + 1 );
			mod->pMaterials[i].pTextureFilename = malloc(strlen( pathTemp )+1);
			strcpy( mod->pMaterials[i].pTextureFilename, pathTemp );
		}
		else {
			// absolute path
			mod->pMaterials[i].pTextureFilename = malloc(strlen( pMaterial->texture )+1);
			strcpy( mod->pMaterials[i].pTextureFilename, pMaterial->texture );
		}
		pPtr += sizeof( MS3DMATERIAL );
	}

	Model_ReloadTextures(mod);

	float animFPS = *( float* )pPtr;
	pPtr += sizeof( float );

	// skip currentTime
	pPtr += sizeof( float );

	int totalFrames = *( int* )pPtr;
	pPtr += sizeof( int );

	mod->totalTime = totalFrames*1000.0/animFPS;

	mod->numJoints = *( word* )pPtr;
	pPtr += sizeof( word );

	mod->pJoints = malloc(sizeof(JOINT) * mod->numJoints);

	const byte *pTempPtr = pPtr;

	JOINTNAMELISTREC *pNameList = malloc(sizeof(JOINTNAMELISTREC) * mod->numJoints);
	for ( i = 0; i < mod->numJoints; i++ )
	{
		MS3DJOINT *pJoint = ( MS3DJOINT* )pTempPtr;
		pTempPtr += sizeof( MS3DJOINT );
		pTempPtr += sizeof( MS3DKEYFRAME )*( pJoint->numRotationKeyframes + pJoint->numTranslationKeyframes );

		pNameList[i].jointIndex = i;
		pNameList[i].pName = pJoint->name;

	}

	for ( i = 0; i < mod->numJoints; i++ )
	{
		MS3DJOINT *pJoint = ( MS3DJOINT* )pPtr;
		pPtr += sizeof( MS3DJOINT );

		int j, parentIndex = -1;
		if ( strlen( pJoint->parentName ) > 0 )
		{
			for ( j = 0; j < mod->numJoints; j++ )
			{
				if ( stricmp( pNameList[j].pName, pJoint->parentName ) == 0 )
				{
					parentIndex = pNameList[j].jointIndex;
					break;
				}
			}
			if ( parentIndex == -1 ) {
				MessageBox(0, "Unable to find parent bone in MS3D file", "Error", MB_OK);
				return FALSE;
			}
		}

		memcpy( mod->pJoints[i].localRotation, pJoint->rotation, sizeof( float )*3 );
		memcpy( mod->pJoints[i].localTranslation, pJoint->translation, sizeof( float )*3 );
		mod->pJoints[i].parent = parentIndex;
		mod->pJoints[i].numRotationKeyframes = pJoint->numRotationKeyframes;
		mod->pJoints[i].pRotationKeyframes = malloc( sizeof(KEYFRAME) * pJoint->numRotationKeyframes);
		mod->pJoints[i].numTranslationKeyframes = pJoint->numTranslationKeyframes;
		mod->pJoints[i].pTranslationKeyframes = malloc( sizeof(KEYFRAME) * pJoint->numTranslationKeyframes);

		for ( j = 0; j < pJoint->numRotationKeyframes; j++ )
		{
			MS3DKEYFRAME *pKeyframe = ( MS3DKEYFRAME* )pPtr;
			pPtr += sizeof( MS3DKEYFRAME );

			Model_SetJointKeyframe(mod, i, j, pKeyframe->time*1000.0f, pKeyframe->parameter, TRUE );
		}

		for ( j = 0; j < pJoint->numTranslationKeyframes; j++ )
		{
			MS3DKEYFRAME *pKeyframe = ( MS3DKEYFRAME* )pPtr;
			pPtr += sizeof( MS3DKEYFRAME );

			Model_SetJointKeyframe(mod, i, j, pKeyframe->time*1000.0f, pKeyframe->parameter, FALSE );
		}
	}
	free(pNameList);

	Model_SetupJoints(mod);
// { end NEW }
	free(pBuffer);

	// Added Brett Porter, 20020504 - a long ago bugfix
	Model_Restart(mod);

	return TRUE;
}
